@@ -4,10 +4,13 @@
  * Default method-functions for CDK objects.
  *
  * $Author: tom $
- * $Date: 2011/05/15 19:29:07 $
- * $Revision: 1.14 $
+ * $Date: 2019/02/20 20:53:53 $
+ * $Revision: 1.22 $
  */
 
+#define MARKUP_CENTER  "<C>"
+#define MARKUP_NEWLINE "<#10>"
+
 /*
  * Set the object's upper-left-corner line-drawing character.
  */
@@ -67,7 +70,7 @@ void setCdkBXattr (CDKOBJS *obj, chtype ch)
 /*
  * This sets the background color of the widget.
  */
-void setCDKObjectBackgroundColor (CDKOBJS *obj, char *color)
+void setCDKObjectBackgroundColor (CDKOBJS *obj, const char *color)
 {
    chtype *holder = 0;
    int junk1, junk2;
@@ -88,10 +91,33 @@ void setCDKObjectBackgroundColor (CDKOBJS *obj, char *color)
    freeChtype (holder);
 }
 
+static char *convert_NL (const char *source)
+{
+   char *target = strdup (source);
+   char *d = target;
+   const char *s = source;
+   size_t need = strlen (MARKUP_NEWLINE);
+
+   while (*s != '\0')
+   {
+      if (!strncmp (s, MARKUP_NEWLINE, need))
+      {
+	 *d++ = '\n';
+	 s += need;
+      }
+      else
+      {
+	 *d++ = *s++;
+      }
+   }
+   *d = '\0';
+   return target;
+}
+
 /*
  * Set the widget's title.
  */
-int setCdkTitle (CDKOBJS *obj, char *title, int boxWidth)
+int setCdkTitle (CDKOBJS *obj, const char *title, int boxWidth)
 {
    if (obj != 0)
    {
@@ -99,17 +125,16 @@ int setCdkTitle (CDKOBJS *obj, char *title, int boxWidth)
 
       if (title != 0)
       {
-	 chtype *holder;
+	 char *retitle = convert_NL (title);
 	 char **temp = 0;
 	 int titleWidth;
 	 int x;
 	 int len;
 	 int align;
-	 int maxWidth = 0;
 
 	 /* We need to split the title on \n. */
-	 temp = CDKsplitString (title, '\n');
-	 obj->titleLines = (int)CDKcountStrings (temp);
+	 temp = CDKsplitString (retitle, '\n');
+	 obj->titleLines = (int)CDKcountStrings ((CDK_CSTRING2)temp);
 
 	 obj->title = typeCallocN (chtype *, obj->titleLines + 1);
 	 obj->titlePos = typeCallocN (int, obj->titleLines + 1);
@@ -117,10 +142,12 @@ int setCdkTitle (CDKOBJS *obj, char *title, int boxWidth)
 
 	 if (boxWidth >= 0)
 	 {
+	    int maxWidth = 0;
+
 	    /* We need to determine the widest title line. */
 	    for (x = 0; x < obj->titleLines; x++)
 	    {
-	       holder = char2Chtype (temp[x], &len, &align);
+	       chtype *holder = char2Chtype (temp[x], &len, &align);
 	       maxWidth = MAXIMUM (maxWidth, len);
 	       freeChtype (holder);
 	    }
@@ -141,11 +168,57 @@ int setCdkTitle (CDKOBJS *obj, char *title, int boxWidth)
 					      obj->titlePos[x]);
 	 }
 	 CDKfreeStrings (temp);
+	 free (retitle);
       }
    }
    return boxWidth;
 }
 
+char *getCdkTitle (CDKOBJS *obj)
+{
+   int x;
+   int pass;
+   char *result = 0;
+   for (pass = 0; pass < 2; ++pass)
+   {
+      unsigned need = obj->titleLines + 1;
+      for (x = 0; x < obj->titleLines; x++)
+      {
+	 char *title = chtype2String (obj->title[x]);
+	 char *check = 0;
+	 char *format = 0;
+	 if (title == 0)
+	    continue;
+	 need += strlen (title);
+	 if ((check = chtype2Char (obj->title[x])) != 0)
+	 {
+	    /* FIXME - we could infer <R> as well */
+	    if (isspace (*check) || obj->titlePos[x] > 0)
+	       format = MARKUP_CENTER;
+	    free (check);
+	 }
+	 if (x)
+	    need += 5;
+	 if (format)
+	    need += strlen (format);
+	 if (pass)
+	 {
+	    if (x)
+	       strcat (result, MARKUP_NEWLINE);
+	    if (format)
+	       strcat (result, format);
+	    strcat (result, title);
+	 }
+	 free (title);
+      }
+      if (!pass)
+      {
+	 result = typeCallocN (char, need);
+      }
+   }
+   return result;
+}
+
 /*
  * Draw the widget's title.
  */
